# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

from flask import Flask
app = Flask("galah.web")

# Hack to work around the destruction of error handlers by Flask's deferred
# processing.
app.logger_name = "nowhere"
app.logger

from galah.base.config import load_config
app.config.update(load_config("web"))

oauth_enabled = bool(
    app.config.get("GOOGLE_SERVERSIDE_ID") and
    app.config.get("GOOGLE_SERVERSIDE_SECRET") and
    app.config.get("GOOGLE_APICLIENT_ID") and
    app.config.get("GOOGLE_APICLIENT_SECRET")
)

import mongoengine
mongoengine.connect(app.config["MONGODB"])

# Plug the auth system into our app
from auth import login_manager
login_manager.setup_app(app)

import views



# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

from flask import Response, request
from galah.web import app, oauth_enabled
from flask.ext.login import current_user
from galah.web.api.commands import api_calls, UserError
from galah.base.crypto.passcrypt import check_seal, deserialize_seal
from galah.db.models import User
from flask.ext.login import login_user
from galah.web.auth import FlaskUser
from galah.web.util import GalahWebAdapter
import requests
import logging

logger = GalahWebAdapter(logging.getLogger("galah.web.views.api"))

def get_many(dictionary, *args):
    return dict((i, dictionary.get(i)) for i in args)

@app.route("/api/login", methods = ["POST"])
def api_login():
    def success(the_user):
        login_user(the_user)
        
        return Response(
            response = "Successfully logged in.",
            headers = {"X-CallSuccess": "True"}
        )

    def failure():
        return Response(
            response = "Incorrect email or password.",
            headers = {"X-CallSuccess": "False"}
        )

    # If the user is trying to login by giving us an access_token they got from
    # signing in through google, validate the token.
    access_token = request.form.get("access_token", None)
    if access_token and oauth_enabled:
        # Ask google to verify that they authed the user.
        req = requests.post(
            "https://www.googleapis.com/oauth2/v1/tokeninfo",
            data = { "access_token": access_token }
        )

        req_json = req.json()

        email = req_json.get("email", "unknown")

        if req.status_code != requests.codes.ok:
            logger.info("Invalid OAuth2 login by %s.", email)

            return failure()

        # Validate client id is matching to avoid confused deputy attack
        if req_json["audience"] != app.config["GOOGLE_APICLIENT_ID"]:
            logger.error(
                "Non-matching audience field detected for user %s.", email
            )

            return failure()

        # Grab the user from the database (here we also check to make sure that
        # this user actually has account on Galah).
        try:
            user = FlaskUser(User.objects.get(email = req_json["email"]))
        except User.DoesNotExist:
            logger.info(
                "User %s signed in via OAuth2 but a Galah account does not "
                "exist for that user.", email
            )

            return failure()

        logger.info("User %s successfully signed in with OAuth2.", email)

        return success(user)
    elif access_token and not oauth_enabled:
        logger.warning("Attempted login via OAuth2 but OAuth2 is not configured.")
        
        return failure()

    # If the user is trying to authenticate through Galah...
    password = request.form.get("password", None)
    email = request.form.get("email", None)
    if email and password:
        try:
            user = FlaskUser(User.objects.get(email = email))
        except User.DoesNotExist:
            logger.info(
                "User %s tried to sign in via internal auth but a Galah "
                "account does not exist for that user.", email
            )

            return failure()

        if check_seal(password, deserialize_seal(str(user.seal))):
            logger.info(
                "User %s succesfully authenticated with internal auth.", email
            )

            return success(user)
        else:
            logger.info(
                "User %s tried to sign in via internal auth but an invalid "
                "password was given.", email
            )

            return failure()
    
    logger.warning("Malformed request.")

    return failure()

@app.route("/api/call", methods = ["POST"])
def api_call():
    try:
        # Make sure we got some data we can work with
        if request.json is None:
            raise UserError("No request data sent.")

        # The top level object must be either a non-empty list or a dictionary
        # with an api_call key. They will have similar information either way
        # however, so here we extract that information.
        if type(request.json) is list and request.json:
            # The API call's name is the first item in the list, so use that
            # to grab the actual APICall object we need.
            api_name = request.json.pop(0)

            api_args = list(request.json)

            api_kwargs = {}
        elif type(request.json) is dict and "api_name" in request.json:
            # Don't let the user insert their own current_user argument
            if "current_user" in request.json:
                raise UserError("You cannot fool the all-knowing Galah.")
                
            # Resolve the name of the API call and retrieve the actual
            # APICall object we need.
            api_name = request.json["api_name"]
            api_args = request.json["args"]

            del request.json["api_name"]
            del request.json["args"]

            api_kwargs = dict(**request.json)
        else:
            logger.warning("Could not parse request.")

            raise UserError("Request data not formed correctly.")

        if api_name in api_calls:
            api_call = api_calls[api_name]
        else:
            raise UserError("%s is not a recognized API call." % api_name)

        logger.info(
            "API call %s with args=%s and kwargs=%s",
            api_name, str(api_args), str(api_kwargs)
        )

        call_result = api_call(current_user, *api_args, **api_kwargs)

        # If an API call returns a tuple, the second item will be headers that
        # should be added onto the response.
        response_headers = {"X-CallSuccess": True}
        if isinstance(call_result, tuple):
            response_text = call_result[0]
            response_headers.update(call_result[1])
        else:
            response_text = call_result

        return Response(
            response = response_text,
            headers = response_headers,
            mimetype = "text/plain"
        )
    except UserError as e:
        logger.info("UserError raised during API call: %s.", str(e))

        return Response(
            response = "Your command cannot be completed as entered: %s" %
                str(e),
            headers = {
                "X-CallSuccess": "False",
                "X-ErrorType": e.__class__.__name__
            },
            mimetype = "text/plain"
        )
    except Exception as e:
        logger.exception("Exception occured while processing API request.")
        
        return Response(
            response =
                "An internal server error occurred processing your request.",
            headers = {
                "X-CallSuccess": "False",
                "X-ErrorType": "Exception"
            },
            mimetype = "text/plain"
        )



# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

import shutil
import subprocess
import tempfile
import os
import datetime
from bson import ObjectId

from galah.db.models import Submission, Archive

# Set up configuration and logging
from galah.base.config import load_config
config = load_config("sisyphus")

import logging
logger = logging.getLogger("galah.sisyphus.tar_bulk_submissions")

def _tar_bulk_submissions(archive_id, requester, assignment, email = ""):
    archive_id = ObjectId(archive_id)

    archive_file = temp_directory = ""

    # Find any expired archives and remove them
    deleted_files = []
    for i in Archive.objects(expires__lt = datetime.datetime.today()):
        deleted_files.append(i.file_location)

        if i.file_location:
            try:
                os.remove(i.file_location)
            except OSError as e:
                logger.warning(
                    "Could not remove expired archive at %s: %s.",
                    i.file_location, str(e)
                )

        i.delete()

    if deleted_files:
        logger.info("Deleted archives %s.", str(deleted_files))

    # This is the archive object we will eventually add to the database
    new_archive = Archive(
        id = archive_id,
        requester = requester,
        archive_type = "assignment_package"
    )

    temp_directory = archive_file = None
    try:
        # Form the query
        query = {"assignment": ObjectId(assignment)}

        # Only mention email in the query if it's not None or the empty
        # string, otherwise mongo will look for submissions that list the
        # user as None or the empty string (which should be exactly none of
        # the submission in the system).
        if email:
            query["user"] = email

        # Grab all the submissions
        submissions = list(Submission.objects(**query))

        if not submissions:
            logger.info("No submissions found matching query.")
            return

        # Organize all the submissions by user name, as this will closely
        # match the structure of the archive we will build.
        submission_map = {}
        for i in submissions:
            if i.user in submission_map:
                submission_map[i.user].append(i)
            else:
                submission_map[i.user] = [i]

        # Create a temporary directory we will create our archive in.
        temp_directory = tempfile.mkdtemp()

        # Create our directory tree. Instead of making new folders for each
        # submission and copying the user's files over however, we will
        # create symlinks to save space and time.
        for user, user_submissions in submission_map.items():
            # Create a directory for the user
            os.makedirs(os.path.join(temp_directory, user))

            # Create symlinks for all his submissions. Each symlink is
            # named after the submission date.
            for i in user_submissions:
                time_stamp = i.timestamp.strftime("%Y-%m-%d-%H-%M-%S")
                symlink_path = \
                    os.path.join(temp_directory, user, time_stamp)

                # In the highly unlikely event that two of the same user's
                # submissions have the same exact time stamp, we'll need to
                # add a marker to the end of the timestamp.
                marker = 0
                while os.path.exists(symlink_path +
                        ("-%d" % marker if marker > 0 else "")):
                    marker += 1

                if marker > 0:
                    symlink_path += "-%d" % marker

                original_path = i.getFilePath()

                # Detect if the submission's files are still on the filesystem
                if os.path.isdir(original_path):
                    # Create a symlink pointing to the actual submission
                    # directory with the name we gnerated
                    os.symlink(original_path, symlink_path)
                else:
                    # Create an empty text file marking the fact that a
                    # submissions existed but is no longer available.
                    open(symlink_path, "w").close()

        # Create the actual archive file.
        # TODO: Create it in galah's /var/ directory
        file_descriptor, archive_file = tempfile.mkstemp(suffix = ".tar.gz")
        os.close(file_descriptor)

        # Run tar and do the actual archiving. Will block until it's finished.
        subprocess.check_call(
            [
                "tar", "--dereference", "--create", "--gzip", "--directory",
                temp_directory, "--file", archive_file
            ] + submission_map.keys()
        )

        new_archive.file_location = archive_file

        new_archive.expires = \
            datetime.datetime.today() + config["TEACHER_ARCHIVE_LIFETIME"]

        new_archive.save(force_insert = True)
    except Exception as e:
        # If we created a temporary archive file we need to delete it.
        new_archive.file_location = None
        if archive_file:
            os.remove(archive_file)

        new_archive.error_string = str(e)
        new_archive.save(force_insert = True)

        raise
    finally:
        if temp_directory:
            shutil.rmtree(temp_directory)


# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

def plural_if(zstring, zcondition):
    """
    Returns zstring pluralized (adds an 's' to the end) if zcondition is True or
    if zcondition is not equal to 1.

    Example usage could be ``plural_if("cow", len(cow_list))``.

    """

    # If they gave us a boolean value, just use that, otherwise, assume the
    # value is some integral type.
    if type(zcondition) is bool:
        plural = zcondition
    else:
        plural = zcondition != 1
        
    return zstring + ("s" if plural else "")

def pretty_list(the_list, conjunction = "and", none_string = "nothing"):
    """
    Returns a grammatically correct string representing the given list. For
    example...

    >>> pretty_list(["John", "Bill", "Stacy"])
    "John, Bill, and Stacy"
    >>> pretty_list(["Bill", "Jorgan"], "or")
    "Bill or Jorgan"
    >>> pretty_list([], none_string = "nobody")
    "nobody"

    """

    the_list = list(the_list)

    if len(the_list) == 0:
        return none_string
    elif len(the_list) == 1:
        return str(the_list[0])
    elif len(the_list) == 2:
        return str(the_list[0]) + " " + conjunction + " " + str(the_list[1])
    else:
        # Add every item except the last two together seperated by commas
        result = ", ".join(the_list[:-2]) + ", "

        # Add the last two items, joined together by a command and the given
        # conjunction
        result += "%s, %s %s" % \
            (str(the_list[-2]), conjunction, str(the_list[-1]))

        return result

def pretty_timedelta(zdelta):
    # We will build our string part by part. All strings in this array will be
    # concatenated with a space as delimiter.
    stringParts = []
    
    if zdelta.days < 0:
        ago = True
        zdelta = -zdelta
    else:
        ago = False
        stringParts.append("in")

    months = abs(zdelta.days) / 30
    hours = zdelta.seconds / (60 * 60)
    minutes = (zdelta.seconds % (60 * 60)) / 60
    seconds = (zdelta.seconds % 60)

    if months == 0 and zdelta.days == 0 and hours == 0 and minutes == 0 and \
            seconds < 10:
        return "just now"
    
    # Add the months part. Because we only approximate the numbers of months,
    # the rest of the numbers (days, hours, etc) won't be exact so we skip the
    # rest of the if statements.
    if months != 0:
        stringParts += ["about", str(months), plural_if("month", months)]
    
    # Add the days part
    if months == 0 and zdelta.days != 0:
        stringParts += [str(zdelta.days), plural_if("day", zdelta.days)]
    
    # Add the hours part
    if months == 0 and hours != 0:
        stringParts += [str(hours), plural_if("hour", hours)]
    
    # Add the minutes part if we're less than 4 hours away
    if months == 0 and minutes != 0 and zdelta.days == 0 and hours < 4:
        stringParts += [str(minutes), plural_if("minute", minutes)]
        
    # Add the seconds part if we're less than 10 minutes away
    if months == 0 and seconds != 0 and zdelta.days == 0 and hours == 0 and minutes < 10:
        stringParts += [str(seconds), plural_if("second", seconds)]
            
    if ago:
        stringParts.append("ago")
    
    return " ".join(stringParts)
    
def pretty_time_distance(za, zb):
    return pretty_timedelta(zb - za)

def pretty_time(ztime):
    return ztime.strftime("%A, %B %d, %Y @ %I:%M %p")



# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

# The actual view
from galah.web import app
from flask import render_template, url_for
from collections import namedtuple
from werkzeug.exceptions import InternalServerError, NotFound

AsciiPiece = namedtuple("AsciiPiece", ["author", "art"])
ascii_art = {
    500: AsciiPiece(
        author = "pb", 
        art = """     __     __------           
  __/o `\ ,~   _~~  . ..   . ..
 ~ -.   ,'   _~-----           
     `\     ~~~--_'__          
       `~-==-~~~~~---          """        
    ),
    404: AsciiPiece(
        author = "Stephen Morgana",
        art = """             /\\              
              .\\\..          
              \\   \\         
              \ (o) /         
              (/    \         
               /\    \        
              ///     \       
             ///|     |       
            ////|     |       
           //////    /        
           |////    /         
          /|////--V/          
         //\//|   |           
     ___////__\___\__________ 
    ()_________'___'_________)"""
    )
}

messages = {
    500: """It appears something broke, I'm going to go fix it... In the meantime <a href="%s">Try Here</a>.""",
    404: """You seem to have landed in the wrong place... <a href="%s">Try Here</a>."""
}

from galah.web.util import GalahWebAdapter
import logging
logger = GalahWebAdapter(logging.getLogger("galah.web.views.error"))

@app.errorhandler(404)
@app.errorhandler(500)
def error(e):
    # Log the error if it's not a 404 or purposeful abort(500).
    if type(e) is not InternalServerError and type(e) is not NotFound:
        logger.exception("An error occurred while rendering a view.")

    code = e.code if hasattr(e, "code") else 500
    
    if code == 500:
        error_description = "500: Internal Server Error"
    else:
        error_description = str(e)
    
    return render_template(
        "error.html",
        error_description = error_description,
        message = messages[code] % "/home", 
        art_piece = ascii_art[code]
    ), code



# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

from mongoengine import *

#def rigidDocument(zdocClass):
    #def validate(self, *zargs, **zkwargs):
        #for k, v in vars(self).items():
            #if k.startswith("_"): continue
            
            #if k not in self._fields:
                #raise ValidationError("Extraneous attributes were found.")
        
        #return super(zdocClass, self).validate(*zargs, **zkwargs)

    #zdocClass.validate = validate
    
    #return zdocClass
        
#@rigidDocument
#class SubTest(EmbeddedDocument):
    #name = StringField(required = True)
    #score = FloatField()
    #max_score = FloatField()
    #message = StringField()
    #messages = ListField(StringField)

    #meta = {
        #"allow_inheritance": False
    #}

#@rigidDocument
#class TestResult(EmbeddedDocument):
    #sub_tests = ListField(EmbeddedDocumentField(SubTest))
    #score = FloatField()
    #max_score = FloatField()
    
    #meta = {
        #"allow_inheritance": False
    #}

import os.path

# Load up the configuration so we know where the submission directory is
from galah.base.config import load_config
config = load_config("global")

class Submission(Document):
    assignment = ObjectIdField(required = True)
    user = StringField(required = True)
    timestamp = DateTimeField(required = True)
    marked_for_grading = BooleanField()
    most_recent = BooleanField()
    
    # Each filename should be a path relative to the root of the archive they
    # uploaded if they uploaded an archive, otherwise each filename should be
    # just the filename. Include extensions.
    uploaded_filenames = ListField(StringField())
    
    meta = {
        "allow_inheritance": False,
        "indexes": [
            {
                "fields": ("user", "assignment", "-most_recent", "-timestamp"),
                "types": False
            }
        ]
    }

    def getFilePath(self):
        return os.path.join(
            config["SUBMISSION_DIRECTORY"],
            str(self.assignment),
            self.user,
            str(self.id)
        )


# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

from mongoengine import *

class Archive(Document):
    requester = StringField(required = True)
    file_location = StringField()
    error_string = StringField()
    expires = DateTimeField()
    archive_type = StringField(choices = ["assignment_package", "single_submission"], required = True)
    

    meta = {
        "allow_inheritance": False
    }




# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

import zmq

context = zmq.Context()

def send_task(heavylifter_host, task_name, *args, **kwargs):
    # We create a new socket and connect each time because sisyphus tasks
    # should not be sent very often, therefore it's not useful to hold an open
    # socket for large lengths of time without sending much of anything.
    socket = context.socket(zmq.REQ)
    socket.connect(heavylifter_host)

    try:
        socket.send_json({
            "task_name": task_name,
            "args": args,
            "kwargs": kwargs
        })

        # Wait for a reply from the sisyphus.
        poller = zmq.Poller()
        poller.register(socket, zmq.POLLIN)
        if poller.poll(2 * 1000):
            reply = socket.recv_json()
        else:
            raise RuntimeError("sisyphus did not respond.")

        if not reply["success"]:
            raise RuntimeError(
                "sisyphus did not accept task.\n\t" + reply["error_string"]
            )
    finally:
        # Forcibly close the socket.
        socket.close(0)

#print send_task("ipc:///tmp/sisyphus.sock", "test_task", "hi")


# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

from galah.web import app
from flask.ext.login import current_user
from galah.web.auth import account_type_required
from galah.db.models import Class, Assignment, Submission
from flask import render_template, request, abort
import datetime
from mongoengine import Q
from galah.web.util import create_time_element, GalahWebAdapter
import logging

logger = \
    GalahWebAdapter(logging.getLogger("galah.web.views.browse_assignments"))

@app.route("/assignments")
@account_type_required(("student", "teacher"))
def browse_assignments():
    # Grab all the current user's classes
    classes = Class.objects(id__in = current_user.classes).only("name")
    
    # Get the current time so we don't have to do it over and over again.
    now = datetime.datetime.today()

    if "show_all" in request.args:
        assignments = list(Assignment.objects(
            Q(for_class__in = current_user.classes) &
            (Q(hide_until = None) | Q(hide_until__lt = now))
        ).only("name", "due", "due_cutoff", "for_class"))
    else:
        assignments = list(Assignment.objects(
            Q(for_class__in = current_user.classes) &
            (Q(due__gt = now - datetime.timedelta(weeks = 1)) |
            Q(due_cutoff__gt = now - datetime.timedelta(weeks = 1))) &
            (Q(hide_until = None) | Q(hide_until__lt = now))
        ).only("name", "due", "due_cutoff", "for_class"))

    assignments = [i for i in assignments if
            not i.hide_until or i.hide_until < now]

    # Get the number of assignments that we could have gotten if we didn't
    # limit based on due date.
    all_assignments_count = Assignment.objects(
        Q(for_class__in = current_user.classes) &
        (Q(hide_until = None) | Q(hide_until__lt = now))
    ).count()

    submissions = list(Submission.objects(
        user = current_user.email,
        assignment__in = [i.id for i in assignments],
        most_recent = True
    ))

    # Add a property to all the assignments so the template can display their
    # respective class easier. Additionally, add a plain text version of the
    # due date
    for i in assignments:
        try:
            i.class_name = next((j.name for j in classes if j.id == i.for_class))
        except StopIteration:
            logger.error(
                "Assignment with id %s references non-existant class with id "
                "%s." % (str(i.id, i.for_class))
            )
            
            i.class_name = "DNE"

        # Figure out the status messages that we want to display to the user.
        submitted = next((j for j in submissions if j.assignment == i.id), None)
        i.status = i.status_color = None
        if submitted:
            i.status = (
                "You made a submission " +
                create_time_element(submitted.timestamp, now)
            )

            i.status_color = "#84B354"
        elif now < i.due:
            i.status = "You have not submitted yet"

            i.status_color = "#877150"
        elif now > i.due and i.due_cutoff and now > i.due_cutoff:
            i.status = "You have not submitted yet, and it is too late to do so"

            i.status_color = "#E9A400"
        elif now > i.due:
            i.status = "You have not submitted yet!"

            i.status_color = "#FB4313"

    # Sort the assignments by due_cutoff or due date if due_cutoff is not
    # assigned.
    assignments.sort(
        key = lambda i: i.due_cutoff if i.due_cutoff else i.due,
        reverse = True
    )

    return render_template(
        "assignments.html",
        assignments = assignments,
        hidden_assignments = -1 if "show_all" in request.args
                else all_assignments_count - len(assignments),
        create_time_element = create_time_element
    )



# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

from galah.web import app
from flask.ext.login import logout_user, current_user
from galah.web.auth import FlaskUser
from flask import redirect, url_for, flash
from galah.web.util import GalahWebAdapter
import logging

logger = GalahWebAdapter(logging.getLogger("galah.web.views.logout"))

@app.route("/logout")
def logout():
    if current_user.is_authenticated():
        # logout_user() cannot fail...
        logger.info("Succesfully logged out.")

        logout_user()
        
        flash("You have successfully logged out. You may now log in as another "
              "user or close this site.", category = "message")
    else:
        logger.info("User tried to log out when not logged in.")

        flash("You were not logged in and you are still not logged in.",
              category = "error")
    
    return redirect(url_for("login"))



# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

from galah.web import app
from flask.ext.login import current_user
from galah.web.auth import FlaskUser
from flask import (
    redirect, url_for, request, abort, current_app, send_file, Response
)
from galah.db.models import Archive
from bson.objectid import ObjectId, InvalidId
from galah.web.util import GalahWebAdapter
import logging

logger = GalahWebAdapter(logging.getLogger("galah.web.views.get_archive"))

@app.route("/archives/<archive_id>")
def get_archive(archive_id):
    archive = None
    try:
        archive = Archive.objects.get(
            id = ObjectId(archive_id),
            archive_type = "assignment_package"
        )
    except InvalidId:
        logger.info("Invalid ID requested.")

        abort(500)
    except Archive.DoesNotExist:
        pass

    # If we can't find the archive return a 404 error.
    if archive is None:
        logger.info("Could not find archive with given ID.")

        abort(404)

    # Ensure that the requesting user has permission to view this archive
    if not current_user.is_authenticated() or \
            archive.requester != current_user.email:
        logger.info(
            "User requested archive they do not have permission to view."
        )

        return current_app.login_manager.unauthorized()

    if archive.error_string:
        logger.info(
            "User requested archive that had an error during creation: %s.",
            archive.error_string
        )

        return Response(
            response = "Internal server error.",
            headers = {
                "X-CallSuccess": "False",
            },
            mimetype = "text/plain"
        )
    else:
        return send_file(archive.file_location)


#!/usr/bin/env python
# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

# Load Galah's configuration.
from galah.base.config import load_config
config = load_config("sisyphus")

# Set up logging
import logging
logger = logging.getLogger("galah.sisyphus")

# Connect to the mongo database
import mongoengine
mongoengine.connect(config["MONGODB"])

# Set up the queue we will store tasks in
from Queue import Queue
from collections import namedtuple
Task = namedtuple("Task", ("name", "args", "kwargs"))
task_queue = Queue()

# Grab all of the tasks we know about
from tasks import task_list

# All ZMQ operations must be done within a context
import zmq
context = zmq.Context()

# Open up a socket we will listen on.
socket = context.socket(zmq.REP)
socket.bind(config["SISYPHUS_ADDRESS"])

import sys
def consumer():
    # Create a new logger for this consumer
    from threading import current_thread
    logger = logging.getLogger("galah.sisyphus." + current_thread().name)
    
    while True:
        task = task_queue.get()

        try:
            task_list[task.name](*task.args, **task.kwargs)
        except Exception as e:
            if type(e) is TypeError and str(e).startswith("%s()" % task.name):
                logger.error("Task with bad parameters: %s", str(task))
            else:
                logger.warning(
                    "Exception in task %s.", task.name,
                    exc_info = sys.exc_info()
                )

def to_task(request):
    # Do very explicit validation on the request so we can give better error
    # messages.
    is_valid = (
        type(request) is dict and
        all(k in request.keys() for k in ("task_name", "args", "kwargs")) and
        isinstance(request["task_name"], basestring) and
        type(request["args"]) is list and
        type(request["kwargs"]) is dict
    )

    if not is_valid:
        raise RuntimeError("Poorly formed request.")

    return Task(
        name = request["task_name"],
        args = request["args"],
        kwargs = request["kwargs"]
    )

from threading import Thread
consumer_thread = Thread(name = "consumer", target = consumer)
consumer_thread.daemon = True
consumer_thread.start()

def main():
    while True:
        task = socket.recv_json()

        # Convert the JSON dict we got into a Task object.
        try:
            task = to_task(task)
        except RuntimeError as e:
            # sisyphus is only exposed to internal components within Galah.
            # This is a very serious error signifying a problem with Galah's
            # logic.
            logger.exception("Error converting request to Task object.")

            socket.send_json({
                "success": False,
                "error_string": str(e)
            })

            continue

        logger.info("Recieved request for task %s.", task.name)

        # Check if this is a recognized command.
        if task.name not in task_list.keys():
            # Same as above, this is a serious error signifying a problem with
            # Galah's logic.
            logger.error("Unknown task '%s'.", task.name)

            socket.send_json({
                "success": False,
                "error_string": "Unknown task '%s'" % task.name
            })

            continue

        # All is good, place the task in the queue
        task_queue.put(task)

        socket.send_json({"success": True})

if __name__ == "__main__":
    main()


# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

## Create the base form ##
from flask import request, url_for, render_template
from flask.ext.wtf import Form, HiddenField
from galah.web.util import is_url_on_site, GalahWebAdapter
import logging

logger = GalahWebAdapter(logging.getLogger("galah.web.views.login"))

class RedirectForm(Form):
    next = HiddenField()

    def __init__(self, *args, **kwargs):
        super(RedirectForm, self).__init__(*args, **kwargs)
        
        if not self.next.data:
            self.next.data = request.args.get("next") or request.referrer
        
        if not is_url_on_site(app, self.next.data):
            self.next.data = ""
    
    @property
    def redirect_target(self):
        return self.next.data

## Create the login form ##
from flask import redirect
from flask.ext.wtf import Form, TextField, PasswordField, validators
from galah.db.models import User

class LoginForm(RedirectForm):
    email = TextField('Email', [validators.Required(), validators.Email()])
    password = PasswordField('Password', [validators.Required()])

# The actual view
from galah.web import app, oauth_enabled
from galah.base.crypto.passcrypt import check_seal, deserialize_seal
from galah.db.models import User
from flask.ext.login import login_user
from galah.web.auth import FlaskUser
from flask import redirect, url_for, flash, request

# Google OAuth2
from oauth2client.client import OAuth2WebServerFlow

# Google OAuth2 flow object to get user's email.
if oauth_enabled:
    flow = OAuth2WebServerFlow(
        client_id=app.config["GOOGLE_SERVERSIDE_ID"],
        client_secret=app.config["GOOGLE_SERVERSIDE_SECRET"],
        scope="https://www.googleapis.com/auth/userinfo.email",
        redirect_uri=app.config["HOST_URL"] + "/oauth2callback"
    )

@app.route("/login/", methods = ["GET", "POST"])
def login():
    form = LoginForm()

    # If the user's input isn't immediately incorrect (validate_on_submit() will
    # not check if the email and password combo is valid, only that it could be
    # valid)
    if form.validate_on_submit():
        # Find the user with the given email
        try:
            user = FlaskUser(User.objects.get(email = form.email.data))
        except User.DoesNotExist:
            user = None
        
        if oauth_enabled and user and not user.seal:
            flash("You must use R'Mail to login.", category = "error")

            logger.info(
                "User %s has tried to log in via internal auth but their "
                "account does not have a seal.", user.email
            )

            return render_template(
                "login.html", form = form, oauth_enabled = oauth_enabled
            )

        # Check if the entered password is correct
        if not user or \
                not check_seal(form.password.data, deserialize_seal(str(user.seal))):
            flash("Incorrect email or password.", category = "error")

            logger.info(
                "User %s has entered an incorrect email/password pair during "
                "internal auth.", form.email.data
            )
        else:
            login_user(user)

            logger.info(
                "User %s has succesfully logged in via internal auth.",
                user.email
            )

            return redirect(form.redirect_target or url_for("browse_assignments"))
    elif oauth_enabled and request.args.get("type") == "rmail":
        # Login using Google OAuth2
        # Step 1 of two-factor authentication: redirect to AUTH url
        auth_uri = flow.step1_get_authorize_url()
        return redirect(auth_uri)

    return render_template(
        "login.html", form = form, oauth_enabled = oauth_enabled
    )

@app.route('/oauth2callback/')
def authenticate_user():
    """
    Authenticate user as logged in after Google OAuth2 sends a callback.
    """
    error = request.args.get('error')
    if error:
        logger.warning("Google sent us an error via OAuth2: %s", error)

        return redirect(url_for('login'))

    # Get OAuth2 authentication code
    code = request.args.get('code')

    # Exchange code for fresh credentials
    credentials = flow.step2_exchange(code)

    # Extract email and email verification
    id_token = credentials.id_token
    email = id_token['email']
    verified_email = id_token['verified_email']

    if verified_email == 'true':
        # Find the user with the given email
        try:
            user = FlaskUser(User.objects.get(email = email))
        except User.DoesNotExist:
            user = None

        if not user:
            flash("A Galah account does not exist for this email.", "error")

            logger.info(
                "User %s has attempted to log in via OAuth2 but an account "
                "does not exist for them.", email
            )
        else:
            login_user(user)

            logger.info(
                "User %s has succesfully logged in via OAuth2.", email
            )

            return redirect(url_for("home"))

    else:
        flash(u'Sorry, we couldn\'t verify your email', 'error')

        logger.info("User %s failed to authenticate with OAuth2.", email)
      
    return redirect(url_for('login'))



# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

# This module should hopefully remain **small**. Use other libraries and such
# if you can. There shouldn't be that much shared code among the different
# views.

from werkzeug.exceptions import NotFound, HTTPException
from flask import request
def is_url_on_site(app, url):
    """
    Determines if *app* can handle *url*, if this function returns False it
    means url points to a resource off-site.
    
    """
    
    try:
        # Will raise an exception if no endpoint exists for the url
        app.create_url_adapter(request).match(url)
    except NotFound:
        return False
    except HTTPException:
        # Any other exceptions are harmless (I think)
        pass
        
    return True

from galah.base.pretty import pretty_time_distance
import datetime
def create_time_element(timestamp, now = None):
    if now is None:
        now = datetime.datetime.today()

    return '<time datetime"%s" title="%s">%s</time>' % (
        timestamp.strftime("%Y-%m%dT%H:%M:%S"),
        timestamp.strftime("%B %d, %Y at %I:%M:%S %p"),
        pretty_time_distance(now, timestamp)
    )

import logging
from flask.ext.login import current_user
from flask import request
class GalahWebAdapter(logging.LoggerAdapter):
    def __init__(self, logger):
        logging.LoggerAdapter.__init__(self, logger, {})

    def process(self, msg, kwargs):
        prefix = "(user=%s;ip=%s;path=%s) " % (
            current_user.email,
            request.remote_addr,
            request.path
        )

        return (prefix + msg, kwargs)


# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

from mongoengine import *
import datetime

class TestDescription(EmbeddedDocument):
    title = StringField(required = True)
    description = StringField()
    
    meta = {
        "allow_inheritance": False
    }

class TestSpecification(EmbeddedDocument):
    test_driver = StringField(required = True)
    timeout = IntField(default = 30)
    actions = ListField(StringField())
    config = DictField()
    
    meta = {
        "allow_inheritance": False
    }

class Assignment(Document):
    name = StringField(required = True)
    due = DateTimeField(required = True)
    due_cutoff = DateTimeField()
    hide_until = DateTimeField(default = datetime.datetime.min, required = True)
    for_class = ObjectIdField(required = True)
    
    # English description of each of the tests
    tests = ListField(EmbeddedDocumentField(TestDescription))
    
    # Rigid specification used by the shepherd to craft a test request
    test_specification = EmbeddedDocumentField(TestSpecification)

    meta = {
        "allow_inheritance": False
    }

    def validate(self):
        if self.due_cutoff and self.due > self.due_cutoff:
            raise ValidationError("due cannot be later than due_cutoff")

        return Document.validate(self)


# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

from galah.web import app
from galah.web.auth import account_type_required
from galah.db.models import Assignment, Submission, Archive
from bson.objectid import ObjectId
from bson.errors import InvalidId
from flask import send_file, abort
from flask.ext.login import current_user
from galah.web.util import GalahWebAdapter
import os.path
import subprocess
import tempfile
import datetime
import logging
import sys

logger = \
    GalahWebAdapter(logging.getLogger("galah.web.views.download_submission"))

@app.route("/assignments/<assignment_id>/<submission_id>/download.tar.gz")
@account_type_required(("student", "teacher"))
def download_submission(assignment_id, submission_id):
    # Figure out which assignment the user asked for.
    try:
        assignment_id = ObjectId(assignment_id)
        assignment = Assignment.objects.get(id = assignment_id)
    except (InvalidId, Assignment.DoesNotExist) as e:
        logger.info("Could not retrieve assignment: %s.", str(e))
        
        abort(404)

    # Figure out which submission the user is trying to download
    try:
        submission_id = ObjectId(submission_id)
        submission = Submission.objects.get(id = submission_id)
    except InvalidId, Submission.DoesNotExist:
        logger.info("Could not retrieve submission: %s.", str(e))

        abort(404)

    # Find any expired archives and remove them
    deleted_files = []
    for i in Archive.objects(expires__lt = datetime.datetime.today()).limit(5):
        if i.file_location:
            deleted_files.append(i.file_location)

            try:
                os.remove(i.file_location)
            except OSError as e:
                logger.warning(
                    "Could not remove expired archive at %s: %s.",
                    i.file_location, str(e)
                )

        i.delete()

    if deleted_files:
        logger.info("Deleted archives %s.", str(deleted_files))

    new_archive = Archive(
        requester = current_user.id,
        archive_type = "single_submission"
    )

    new_archive.expires = \
        datetime.datetime.today() + app.config["STUDENT_ARCHIVE_LIFETIME"]

    archive_file_name = ""
    try:
        # Create the actual archive file.
        # TODO: Create it in galah's /var/ directory
        archive_fd, archive_file_name = tempfile.mkstemp(suffix = ".tgz")

        # Close the file because we won't actually do any writing to it,
        # rather tar will.
        os.close(archive_fd)

        # Run tar and do the actual archiving. Will block until it's finished.
        subprocess.check_call(
            [
                "tar", "--dereference", "--create", "--gzip", "--directory",
                submission.getFilePath(),
                "--file", archive_file_name
            ] + submission.uploaded_filenames
        )

        new_archive.file_location = archive_file_name

        new_archive.save(force_insert = True)

        return send_file(archive_file_name)
    except Exception as e:
        logger.exception("An error occured while creating an archive.")

        # If we created a temporary archive file we need to delete it.
        new_archive.file_location = None
        if archive_file_name:
            os.remove(archive_file_name)

        new_archive.error_string = str(e)
        new_archive.save(force_insert = True)

        abort(500)


# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

# Wrapper around User to allow Flask-Login to work with it
from types import MethodType
def FlaskUser(user):
    "Decorates a given User instance with the methods in the UserMixin class"
    
    methodize = lambda func: MethodType(func, user)
    
    # Too many problems were occuring transplanting the UserMixin class so I
    # remade the methods here.
    user.is_active = methodize(lambda self: True)
    user.is_authenticated = methodize(lambda self: True)
    user.is_anonymous = methodize(lambda self: False)
    user.get_id = methodize(lambda self: self.id)
    
    return user

# All users must have an account_type variable in order for the 
# account_type_required decorator to work
from flask.ext.login import AnonymousUser
class Anonymous(AnonymousUser):
    account_type = None
    email = "unknown"

## Decorator that enforces account_type based permissions ##
from functools import wraps
from flask.ext.login import current_user
from flask import current_app, flash
from types import StringType
from galah.base.pretty import pretty_list
from galah.web import app
def account_type_required(account_type):
    """
    A decorator that can be applied to views to only allow access to users with
    certain account types.

    :param account_type: The account types that are allowed. May either be a
                         single string (meaning only one account type is
                         allowed) or a tuple, in which case any account type
                         within that tuple will be allowed access.

    """

    # We allow the user of this function to specify a string for account_type
    # signaling that only one type of user is allowed. The rest of the function
    # except account_type to be a tuple however, so convert it here.
    if isinstance(account_type, StringType):
        account_type = (account_type, )

    # Form a nicely formatted string that we will use to provide an error
    # message to the end-user if they try to access a restricted page.
    allowed = pretty_list(account_type, none_string = "magical")

    def internal_decorator(func):
        @wraps(func)
        def decorated_view(*args, **kwargs):
            if not current_user.is_authenticated() or \
                    current_user.account_type not in account_type:
                flash("Only %s users are permitted to access this page."
                          % allowed, category = "error")

                return current_app.login_manager.unauthorized()
                
            return func(*args, **kwargs)
        
        return decorated_view
        
    return internal_decorator

# Set up the login manager
from flask.ext.login import LoginManager
login_manager = LoginManager()
login_manager.anonymous_user = Anonymous
login_manager.login_view = "login"
login_manager.login_message = None
    
# Loads a user useable by Flask-Login from a user id (email)
from galah.db.models import User
@login_manager.user_loader
def load_user(email):
    return FlaskUser(User.objects.get(email = email))



#!/usr/bin/env python

# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

def create_admin(email, password, galah_server = "galah"):
    # First we need to get mongoengine connected to our mongo installation
    import mongoengine
    mongoengine.connect(galah_server)

    # Then we will access the create_user API command directly, and use the
    # dummy user "admin_user" to fake a current user
    from galah.web.api.commands import create_user, admin_user
    create_user(admin_user, email, password, "admin")

import sys
def parse_arguments(args = sys.argv[1:]):
    from optparse import OptionParser, make_option

    option_list = [
        make_option(
            "--password-env", "-e", metavar = "VARIABLE_NAME",
            dest = "password_env",
            help = "The name of an environmental variable to pull the password "
                   "for the new admin user from."
        )
    ]

    parser = OptionParser(
        usage = "Usage: %prog [OPTIONS] EMAIL [PASSWORD]",
        description = "Command line interface to Galah for use by instructors "
                      "and administrators. If PASSWORD is not specified and "
                      "-p is not specified, user will be prompted for the "
                      "password.",
        option_list = option_list
    )

    options, pos_args = parser.parse_args(args)

    if len(pos_args) != 1 and len(pos_args) != 2:
        parser.error(
            "Exactly 1 or 2 positional arguments may be supplied, %d given." %
                len(pos_args)
        )

    return (options, pos_args)

def main():
    options, args = parse_arguments()

    if len(args) == 2:
        password = args[1]
    elif options.password_env:
        import os
        password = os.environ.get(options.password_env)

        if not password:
            print >> sys.stderr, (
                "Could not get password from environmental variable %s." %
                    options.password_env
            )

            return 1
    else:
        # Grab the password from the user.
        import getpass
        password = getpass.getpass("What should this user's password be? ")

        if not password:
            print >> sys.stderr, "No password given. Aborting."

            return 1

    print "Creating admin user", args[0]
    try:
        create_admin(args[0], password)
    except Exception as e:
        print >> sys.stderr, "Could not create admin user.", str(e)

        return 1

    return 0

if __name__ == "__main__":
    exit(main())



# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

## The Actual View ##
from galah.web import app
from flask.ext.login import current_user
from galah.web.auth import account_type_required
from bson.objectid import ObjectId
from bson.errors import InvalidId
from flask import abort, render_template, request, flash, redirect, jsonify, \
                  url_for
from galah.db.models import Submission, Assignment
from galah.base.pretty import pretty_list, plural_if
from galah.web.util import is_url_on_site, GalahWebAdapter
from werkzeug import secure_filename
import os.path
import subprocess
import datetime
import shutil
import tempfile
import datetime
import logging

from _view_assignment import SimpleArchiveForm

logger = \
    GalahWebAdapter(logging.getLogger("galah.web.views.upload_submissions"))

@app.route("/assignments/<assignment_id>/upload", methods = ["POST"])
@account_type_required(("student", "teacher"))
def upload_submission(assignment_id):
    # Figure out which assignment the user asked for.
    try:
        id = ObjectId(assignment_id)
        assignment = Assignment.objects.get(id = id)
    except (InvalidId, Assignment.DoesNotExist) as e:
        logger.info("Could not retrieve assignment: %s", str(e))
        
        abort(404)
    
    # Figure out where we should redirect the user to once we're done.
    redirect_to = request.args.get("next") or request.referrer

    if not is_url_on_site(app, redirect_to):
        # Default going back to the assignment screen
        redirect_to = url_for(
            "view_assignment",
            assignment_id = assignment_id
        )
    
    # Check if the assignment's cutoff date has passed
    if assignment.due_cutoff and \
            assignment.due_cutoff < datetime.datetime.today():
        logger.info("Submission rejected, cutoff date has already passed.")

        return craft_response(
            error = "The cutoff date has already passed, your submission was "
                    "not accepted."
        )

    form = SimpleArchiveForm()
    if not form.validate_on_submit():
        logger.info("Submission rejected. Validation failure.")

        flash("The files you passed in were invalid.", category = "error")

        return redirect(redirect_to)

    if not [i for i in form.archive.entries if i.data.filename]:
        logger.info("Submission rejected. User did not submit any files.")

        flash("You did not submit any files.", category = "error")

        return redirect(redirect_to)

    new_submission = Submission(
        assignment = id,
        user = current_user.id,
        timestamp = datetime.datetime.now(),
        marked_for_grading = True,
        most_recent = True
    )
    new_submission.id = ObjectId()

    # Craft a unique directory path where we will store the new submission. We
    # are guarenteed an ObjectId is unique. However we are not guarenteed that
    # we will have the proper permissions and that we will be able to make the
    # directory thus this could error because of that.
    new_submission.testables = new_submission.getFilePath()
    os.makedirs(new_submission.testables)

    # Save each file the user uploaded into the submissions directory
    for i in form.archive.entries:
        if not i.data.filename:
            continue

        # Figure out where we want to save the user's file
        file_path = os.path.join(
            new_submission.testables, secure_filename(i.data.filename)
        )

        # Do the actual saving
        i.data.save(file_path)
    
    new_submission.uploaded_filenames.extend(
        secure_filename(i.data.filename) for i in form.archive.entries
            if i.data.filename
    )

    logger.info(
        "Succesfully uploaded a new submission (id = %s) with files %s.",
        str(new_submission.id),
        str(new_submission.uploaded_filenames)
    )

    # The old "most_recent" submission is no longer the most recent.
    Submission.objects(
        user = current_user.email, 
        assignment = id, 
        most_recent = True
    ).update(
        multi = False,
        unset__most_recent = 1
    )

    new_submission.save()
    
    # Communicate to the next page what submission was just added.
    flash(new_submission.id, category = "new_submission")

    flash(
        "Successfully uploaded %s %s." %
            (
                plural_if("file", len(new_submission.uploaded_filenames)),
                pretty_list(new_submission.uploaded_filenames)
            ),
        category = "message"
    )
    
    # Everything seems to have gone well
    return redirect(redirect_to)



# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

from collections import namedtuple
import os
import pbkdf2

# The known algorithms
algorithms = ("pbkdf2", )

# A hash along with all the information used to create it
HashSeal = namedtuple("HashSeal", ("hash", "algorithm", "salt", "cost_factor"))

def hash(password, algorithm, salt, cost_factor, **kwargs):
    """
    Returns a hashed password (not a HashSeal) in binary format.
    
    **kwargs are passed to the underlying hash function unchanged.
    
    """
    
    if algorithm == "pbkdf2":
        return pbkdf2.pbkdf2_bin(str(password), salt, cost_factor, **kwargs)
    else:
        raise ValueError("algorithm: Specified algorithm is not recognized.")

def seal(password, algorithm = "pbkdf2", salt = None, cost_factor = 1000):
    "Returns a HashSeal of the given password."
    
    if salt is None:
        salt = os.urandom(4)
        
    return HashSeal(
        hash = hash(password, algorithm, salt, cost_factor),
        algorithm = algorithm,
        salt = salt,
        cost_factor = cost_factor
    )
    
def check_seal(password, seal):
    "Returns True if a password is the password used to create the given seal."
    
    def safeStrCmp(a, b):
        if len(a) != len(b):
            return False
        
        rv = True
        for x, y in zip(a, b):
            if x != y:
                rv = False
        
        return rv
    
    # Create the hash of the password we will check against the seal. Note its
    # important that password is a string (not unicode) because ord() needs to
    # work with it.
    check_hash = hash(password, seal.algorithm, seal.salt, seal.cost_factor)
    
    return safeStrCmp(check_hash, seal.hash)
    
def serialize_seal(seal):
    return ";".join((
        seal.hash.encode("hex"), 
        seal.algorithm, 
        seal.salt.encode("hex"), 
        str(seal.cost_factor)
    ))

def deserialize_seal(seal):
    parts = seal.split(";")
    parts[0] = str(parts[0].decode("hex"))
    parts[1] = unicode(parts[1])
    parts[2] = str(parts[2].decode("hex"))
    parts[3] = int(parts[3])
    
    return HashSeal(*parts)

if __name__ == "__main__":
    test_seal = seal("test")
    print test_seal
    
    assert check_seal("test", test_seal)
    
    print serialize_seal(test_seal), deserialize_seal(serialize_seal(test_seal))
    assert deserialize_seal(serialize_seal(test_seal)) == test_seal



# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

import datetime

# Default configuration values for Galah.
defaults = {
    "global/CURRENT_VERSION": "v0.1.3",
    "global/SUBMISSION_DIRECTORY": "/var/local/galah/web/submissions/",
    "global/MONGODB": "galah",
    "global/SISYPHUS_ADDRESS": "ipc:///tmp/sisyphus.sock",
    "web/DEBUG": True,
    "web/SECRET_KEY": "Very Secure Key",
    "web/HOST_URL": "http://localhost:5000",
    "web/STUDENT_ARCHIVE_LIFETIME": datetime.timedelta(minutes = 2),
    "web/SOURCE_HOST": "https://github.com/brownhead/galah",
    "sisyphus/TEACHER_ARCHIVE_LIFETIME": datetime.timedelta(minutes = 2)
}

import imp
import logging
import os
from galah.base.utility import tuplify

loaded = None
def load_config(domain):
    """
    Parses and loads up the Galah configuration file and extracts the
    configuration data for the given domain (ex: "web" or "sheep").

    Global configuration options will always be returned, no matter the domain.

    The configuration file will only be loaded once per instance of the
    interpreter, so feel free to call this function multiple times in many files
    to avoid ugly dependency issues.

    """

    global loaded

    # Note whether this is the first time we're processing the configuraiton. If
    # it is, we'll want to attach the global log handlers.
    first_load = loaded is None

    # Load the configuration if it hasn't been loaded yet. Note this ensures
    # that the configuration is only loaded once per instance of the
    # interpreter
    if not loaded:
        if os.path.isfile("/etc/galah/galah.config"):
            loaded = imp.load_source(
                "user_config_file", "/etc/galah/galah.config"
            )

    local_config = {}

    # Note we make a **shallow** copy of the defaults here
    user_config = dict(defaults)

    # If the configuration didn't load correctly, just use the defaults
    if loaded:
        user_config.update(loaded.config)

    # The prefix values we will look for when scanning the configuration file.
    prefix = "%s/" % domain
    global_prefix = "global/"

    # Grab all the configuration values for the given domain, and any global
    # configuration values.
    for k, v in user_config.items():
        if k.startswith(prefix) and k != len(prefix):
            local_key = k[len(prefix):]

            local_config[local_key] = v
        elif k.startswith(global_prefix) and k != len(global_prefix):
            global_key = k[len(global_prefix):]

            local_config[global_key] = v

    return local_config



#!/usr/bin/env python

# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

def main():
    # FIgure out what to delete.
    delete_submissions = raw_input(
        "Are you sure you want to delete all submissions? ") == "yes"
    delete_assignments = delete_submissions and raw_input(
        "Are you sure you want to delete all assignments? ") == "yes"
    delete_users = delete_assignments and raw_input(
        "Are you sure you want to delete all non-admin users? ") == "yes"
    delete_classes = delete_users and raw_input(
        "Are you sure you want to delete all classes? ") == "yes"

    # First we need to get mongoengine connected to our mongo installation
    import mongoengine
    mongoengine.connect("galah")

    from galah.base.config import load_config
    config = load_config("global")

    if delete_submissions:
        # Delete all of the submissions from the filesystem
        import subprocess
        print "Deleting all submissions from the filesystem..."
        subprocess.check_call(["rm", "-r", config["SUBMISSION_DIRECTORY"]])
    
        import galah.db.models as models

        # Delete all of the submissions from the database. Submissions point to
        # both users and assignments, so they must be deleted before both.
        print "Deleting all submissions from the database..."
        models.Submission.objects().delete()

    if delete_users:
        # Delete all of the non-admin users from the database. Users point to
        # classes.
        print "Deleting all non-admin users from the database..."
        models.User.objects(account_type__ne = "admin").delete()

    if delete_assignments:
        # Delete all of the assignments from the database. Assignments point to
        # classes.
        print "Deleting all assignments from the database..."
        models.Assignment.objects().delete()

    if delete_classes:
        # Delete all of the classes from the database.
        print "Deleting all classes from the database..."
        models.Class.objects().delete()

if __name__ == "__main__":
    main()



# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

from inspect import getargspec
from copy import deepcopy
from bson import ObjectId
from bson.errors import InvalidId
from galah.db.models import *
import json
from collections import namedtuple
from mongoengine import ValidationError
from galah.sisyphus.api import send_task

from galah.base.config import load_config
config = load_config("web")

#: An anonymous admin user useful when accessing this module from the local
#: system.
admin_user = namedtuple("User", "account_type")("admin")

class UserError(Exception):
    def __init__(self, what):
        self.what = str(what)

    def __str__(self):
        return self.what

class PermissionError(UserError):
    def __init__(self, *args, **kwargs):
        UserError.__init__(self, *args, **kwargs)

class APICall(object):
    """Wraps an API call and handles basic permissions along with providing a
    simple interface to get meta data on the API call.
    
    """
    
    __slots__ = ("wrapped_function", "allowed", "argspec", "name")
    
    def __init__(self, wrapped_function, allowed = None):
        #: The raw function we are wrapping that performs the actual logic.
        self.wrapped_function = wrapped_function
        
        #: The account types that are allowed to call this function, if None 
        #: any account type may call this function.
        self.allowed = allowed
        
        #: Information about the arguments this API call accepts in the same
        #: format :func: inspect.getargspec returns.
        self.argspec = getargspec(wrapped_function)
        
        #: The name of the wrapped function.
        self.name = wrapped_function.func_name
        
    def __call__(self, current_user, *args, **kwargs):
        arg_spec = getargspec(self.wrapped_function)
        
        # If no validation is required this won't actually be a problem, however
        # it's certainly not something that you should be doing.
        if not hasattr(current_user, "account_type"):
            raise ValueError("current_user is not a valid user object.")
        
        # Check if the current user has permisson to perform this operation
        if self.allowed and current_user.account_type not in self.allowed:
            raise PermissionError(
                "Only %s users are allowed to call %s" %
                    (
                        " or ".join(self.allowed),
                        self.wrapped_function.func_name
                    )
            )
        
        # Only pass the current user to the function if the function wants it
        if len(self.argspec[0]) != 0 and self.argspec[0][0] == "current_user":
            return self.wrapped_function(current_user, *args, **kwargs)
        else: 
            return self.wrapped_function(*args, **kwargs)

from decorator import decorator

def _api_call(allowed = None):
    """Decorator that wraps a function with the :class: APICall class."""
    
    if isinstance(allowed, basestring):
        allowed = (allowed, )

    def inner(func):
        return APICall(func, allowed)

    return inner

## Some useful low level functions ##
def _get_user(email, current_user):
    if email == "me":
        return current_user

    try:
        return User.objects.get(email = email)
    except User.DoesNotExist:
        raise UserError("User %s does not exist." % email)

def _user_to_str(user):
    return "User [email = %s, account_type = %s]" % \
            (user.email, user.account_type)

def _get_assignment(query, current_user):
    # Check if class/assignment syntax was used.
    if "/" in query:
        parts = query.split("/", 1)

        if parts[0] == "mine":
            class_string = "classes you are enrolled in or assigned to"

            matches = list(Assignment.objects(
                for_class__in = current_user.classes,
                name__icontains = parts[1]
            ))
        else:
            the_class = _get_class(parts[0])

            class_string = _class_to_str(the_class)

            matches = list(Assignment.objects(
                for_class = the_class.id,
                name__icontains = parts[1]
            ))

        if not matches:
            raise UserError(
                "No assignments in %s matched your query of {name "
                "contains '%s'}." % (class_string, parts[1])
            )
        elif len(matches) == 1:
            return matches[0]
        else:
            raise UserError(
                "%d assignments in %s matched your query of {name "
                "contains '%s'}, however this API expects 1 assignment. Refine "
                "your query and try again.\n\t%s" % (
                    len(matches),
                    class_string,
                    parts[1],
                    "\n\t".join(_assignment_to_str(i) for i in matches)
                )
            )

    try:
        return Assignment.objects.get(id = ObjectId(query))
    except Assignment.DoesNotExist:
        raise UserError("Assignment with ID %s does not exist." % query)
    except InvalidId:
        raise UserError("Assignment ID %s is not a valid ID." % query)

def _assignment_to_str(assignment):
    return "Assignment [id = %s, name = %s]" % (assignment.id, assignment.name)

def _get_class(query, instructor = None):
    try:
        query_dict = {"id": ObjectId(query)}
        if instructor:
            query_dict["id__in"] = instructor.classes

        # Check if the user provided a valid ObjectId
        return Class.objects.get(**query_dict)
    except (Class.DoesNotExist, InvalidId):
        pass

    query_dict = {"name__icontains": query}
    if instructor:
        query_dict["id__in": instructor.classes]

    matches = list(Class.objects(**query_dict))
    
    if not matches:
        raise UserError("No classes matched your query of '%s'." % query)    
    elif len(matches) == 1:
        return matches[0]
    else:
        raise UserError(
            "%d classes match your query of '%s', however, this API expects 1 "
            "class. Refine your query and try again.\n\t%s" % (
                len(matches),
                query, 
                "\n\t".join(_class_to_str(i) for i in matches)
            )
        )

def _class_to_str(the_class):
    return "Class [id = %s, name = %s]" % (the_class.id, the_class.name)
        
import datetime
def _to_datetime(time):
    try:
        return datetime.datetime(time)
    except TypeError:
        pass
        
    try:
        return datetime.datetime.strptime(time, "%m/%d/%Y %H:%M:%S")
    except (OverflowError, ValueError):
        raise ValueError(
            "Could not convert %s into a time object." % repr(time)
        )

def _datetime_to_str(time):
    try:
        return datetime.datetime.strftime(time, "%m/%d/%Y %H:%M:%S")
    except TypeError:
        return "None"
        
## Below are the actual API calls ##
@_api_call()
def get_api_info():
    # TODO: This function should be memoized
    
    api_info = []
    for k, v in api_calls.items():
        api_info.append({"name": k})
        current = api_info[-1]
        
        # Loop through all the arguments the function takes in and add
        # information on each argument to the api_info
        current["args"] = []
        for i in xrange(len(v.argspec.args)):
            current["args"].append({"name": v.argspec.args[i]})
            
            if v.argspec.defaults:
                # The number of arguments without default values
                ndefaultless = len(v.argspec.args) - len(v.argspec.defaults)
                
                # If the current argument has a default value make note of it
                if ndefaultless <= i:
                    current["args"][-1].update({
                        "default_value": v.argspec.defaults[i - ndefaultless]
                    })
    
    return json.dumps(api_info, separators = (",", ":"))

@_api_call()
def whoami(current_user):
    if hasattr(current_user, "id"):
        return current_user.id
    else:
        return "Anonymous"

@_api_call()
def get_oauth2_keys():
    from galah.web import app
    import json

    google_api_keys = {
        "CLIENT_ID": app.config["GOOGLE_APICLIENT_ID"],
        "CLIENT_SECRET": app.config["GOOGLE_APICLIENT_SECRET"]
    }

    return json.dumps(google_api_keys, separators = (",", ":"))

from galah.base.crypto.passcrypt import serialize_seal, seal
from mongoengine import OperationError
@_api_call("admin")
def create_user(email, password = "", account_type = "student"):
    new_user = User(
        email = email, 
        account_type = account_type
    )

    if password:
        new_user.seal = serialize_seal(seal(password))
    
    try:
        new_user.save(force_insert = True)
    except OperationError:
        raise UserError("A user with that email already exists.")
    
    return "Success! %s created." % _user_to_str(new_user)

@_api_call("admin")
def reset_password(current_user, email, new_password = ""):
    the_user = _get_user(email, current_user)

    if new_password:
        the_user.seal = serialize_seal(seal(new_password))
    else:
        the_user.seal = None

    the_user.save()

    return "Success! Password for %s succesfully reset." \
                % _user_to_str(the_user)

@_api_call("admin")
def modify_user(current_user, email, account_type):
    the_user = _get_user(email, current_user)

    old_user_string = _user_to_str(the_user)

    the_user.account_type = account_type

    try:
        the_user.save()
    except ValidationError:
        raise UserError("%s is not a valid account type." % account_type)

    return "Success! %s has been retyped as a %s" \
                % (old_user_string, account_type)

@_api_call(("admin", "teacher"))
def find_user(current_user, email_contains = "", account_type = "",
              enrolled_in = "", max_results = "20"):
    max_results = int(max_results)

    query = {}
    query_description = []

    if email_contains:
        query["email__icontains"] = email_contains
        query_description.append("email contains '%s'" % email_contains)

    if account_type:
        query["account_type"] = account_type
        query_description.append("account type is '%s'" % account_type)

    if enrolled_in:
        the_class = _get_class(enrolled_in)
        query["classes"] = the_class.id
        query_description.append("enrolled in %s" %
                _class_to_str(the_class))

    matches = list(User.objects(**query)[:max_results + 1])

    # Check if there are more than max_results results
    plus = ""
    if len(matches) > max_results:
        matches.pop()
        plus = "+"

    if query_description:
        query_description = ",".join(query_description)
    else:
        query_description = "any"

    result_string = "\n\t".join(_user_to_str(i) for i in matches)

    return "%d%s user(s) found matching query {%s}.\n\t%s" % \
            (len(matches), plus, query_description, result_string)

@_api_call(("admin", "teacher"))
def user_info(current_user, email):
    user = _get_user(email, current_user)

    enrolled_in = Class.objects(id__in = user.classes)

    class_list = "\n\t".join(_class_to_str(i) for i in enrolled_in)

    if not enrolled_in:
        return "%s is not enrolled in any classes." % _user_to_str(user)
    else:
        return "%s is enrolled in:\n\t%s" % (_user_to_str(user), class_list)

@_api_call("admin")
def delete_user(current_user, email):    
    to_delete = _get_user(email, current_user)

    to_delete.delete()
        
    return "Success! %s deleted." % _user_to_str(to_delete)
    
@_api_call(("admin", "teacher"))
def find_class(current_user, name_contains = "", enrollee = ""):
    if not enrollee and current_user.account_type != "admin":
        query = {"id__in": current_user.classes}
        instructor_string = "You are"
    elif enrollee == "any" or current_user.account_type == "admin":
        query = {}
        instructor_string = "Anyone is"
    else:
        the_instructor = _get_user(enrollee, current_user)
        query = {"id__in": the_instructor.classes}
        instructor_string = _user_to_str(the_instructor) + " is"
        
    if name_contains:
        query["name__icontains"] = name_contains

    matches = list(Class.objects(**query))

    result_string = "\n\t".join(_class_to_str(i) for i in matches)

    return "%s teaching %d class(es) with '%s' in their name.\n\t%s" % \
            (instructor_string, len(matches), name_contains, result_string)

@_api_call(("admin", "teacher"))
def enroll_student(current_user, email, enroll_in):
    user = _get_user(email, current_user)

    if current_user.account_type != "admin" and \
            user.account_type == "teacher":
        raise PermissionError("Only admins can assign teachers to classes.")
    
    the_class = _get_class(enroll_in)

    if the_class.id in user.classes:
        raise UserError("%s is already enrolled in %s." %
            (_user_to_str(user), _class_to_str(the_class))
        )
            
    user.classes.append(the_class.id)
    user.save()
    
    return "Success! %s enrolled in %s." % (
        _user_to_str(user), _class_to_str(the_class)
    )

@_api_call("admin")
def assign_teacher(current_user, email, assign_to):
    return enroll_student(current_user, email, assign_to)

@_api_call(("admin", "teacher"))
def drop_student(current_user, email, drop_from):    
    if current_user.account_type == "admin":
        the_class = _get_class(drop_from)
    else:
        the_class = _get_class(drop_from, current_user)

    user = _get_user(email, current_user)
        
    if the_class.id not in user.classes:
        raise UserError("%s is not enrolled in %s." %
            (_user_to_str(user), _class_to_str(the_class))
        )
    
    user.classes.remove(the_class.id)
    user.save()
    
    return "Success! Dropped %s from %s." % (
        _user_to_str(user), _class_to_str(the_class)
    )

@_api_call(("admin", "teacher"))
def unassign_teacher(current_user, email, drop_from):
    return drop_student(current_user, email, drop_from)

@_api_call(("admin", "teacher"))
def class_info(for_class):
    the_class = _get_class(for_class)

    assignments = Assignment.objects(for_class = the_class.id)
    if assignments:
        assignments_string = "\n\t".join(
            _assignment_to_str(i) for i in assignments
        )
    else:
        assignments_string = "(No assignments)"

    return _class_to_str(the_class) + " has assignments:\n\t" + \
            assignments_string

@_api_call("admin")
def create_class(name):
    new_class = Class(name = name)
    new_class.save()
    
    return "Success! %s created." % _class_to_str(new_class)

@_api_call("admin")
def modify_class(the_class, name):
    if not name:
        raise UserError("name cannot be empty.")

    the_class = _get_class(the_class)

    old_class_string = _class_to_str(the_class)

    the_class.name = name

    the_class.save()

    return "Success! %s has been renamed to '%s'" % (old_class_string, name)

@_api_call("admin")
def delete_class(to_delete):
    the_class = _get_class(to_delete)
        
    # Get all of the assignments for the class
    assignments = \
        [str(i.id) for i in Assignment.objects(for_class = the_class.id)]
    
    send_task(
        config["SISYPHUS_ADDRESS"],
        "delete_assignments",
        assignments,
        str(the_class.id)
    )

    return (
        "%s has been queued for deletion. Please allow a few minutes for the "
        "task to complete." % _class_to_str(the_class)
    )

@_api_call(("admin", "teacher"))
def create_assignment(current_user, name, due, for_class, due_cutoff = "",
                      hide_until = ""):
    # The attributes of the assignmnet we're creating
    atts = {"name": name}

    atts["due"] = _to_datetime(due)

    if due_cutoff:
        atts["due_cutoff"] = _to_datetime(due_cutoff)

    if hide_until:
        atts["hide_until"] = _to_datetime(hide_until)

    the_class = _get_class(for_class)

    if current_user.account_type != "admin" and \
            the_class.id not in current_user.classes:
        raise PermissionError(
            "You cannot create an assignment for a class you are not teaching."
        )

    atts["for_class"] = the_class.id

    print atts
    new_assignment = Assignment(**atts)
    new_assignment.save()
    
    return "Success! %s created." % _assignment_to_str(new_assignment)

@_api_call(("admin", "teacher"))
def assignment_info(current_user, id):
    assignment = _get_assignment(id, current_user)

    attribute_strings = []
    for k, v in assignment._data.items():
        if k and v:
            if type(v) is datetime.datetime:
                if v == datetime.datetime.min:
                    continue

                v = _datetime_to_str(v)

            attribute_strings.append("%s = %s" % (k, v))

    attributes = "\n\t".join(attribute_strings)

    return "Properties of %s:\n\t%s" \
                % (_assignment_to_str(assignment), attributes)

@_api_call(("admin", "teacher"))
def modify_assignment(current_user, id, name = "", due = "", for_class = "",
                      due_cutoff = "", hide_until = ""):
    assignment = _get_assignment(id, current_user)

    # Save the string representation of the original assignment show we can show
    # it to the user later.
    old_assignment_string = _assignment_to_str(assignment)

    if current_user.account_type != "admin" and \
            assignment.for_class not in current_user.classes:
        raise PermissionError(
            "You can only modify assignments for classes you teach."
        )

    change_log = []

    if name:
        change_log.append(
            "Name changed from '%s' to '%s'." % (assignment.name, name)
        )

        assignment.name = name

    if due:
        due_date = _to_datetime(due)

        change_log.append(
            "Due date changed from '%s' to '%s'."
                % (_datetime_to_str(assignment.due), _datetime_to_str(due_date))
        )

        assignment.due = due_date

    if due_cutoff:
        if due_cutoff.lower() == "none":
            cutoff_date = None
        else:
            cutoff_date = _to_datetime(due_cutoff)

        change_log.append(
            "Cutoff date changed from '%s' to '%s'."
                % (
                    _datetime_to_str(assignment.due_cutoff),
                    _datetime_to_str(cutoff_date)
                )
        )

        assignment.due_cutoff = cutoff_date

    if hide_until:    
        if hide_until.lower() == "none":
            hide_until = datetime.datetime.min
        else:
            hide_until = _to_datetime(hide_until)

        change_log.append(
            "Hide-until date changed from '%s' to '%s'."
                % (str(assignment.hide_until), str(hide_until))
        )

        assignment.hide_until = hide_until

    if for_class:
        old_class = Class.objects.get(id = ObjectId(assignment.for_class))
        new_class = _get_class(for_class)

        change_log.append(
            "Class changed from %s to %s."
                % (_class_to_str(old_class), _class_to_str(new_class))
        )

        assignment.for_class = new_class.id

        if current_user.account_type != "admin" and \
                assignment.for_class not in current_user.classes:
            raise PermissionError(
                "You cannot reassign an assignment to a class you're not "
                "teaching."
            )

    assignment.save()

    if change_log:
        change_log_string = "\n\t".join(change_log)
    else:
        change_log_string = "(No changes)"

    return "Success! The following changes were applied to %s.\n\t%s" \
                % (old_assignment_string, change_log_string)

@_api_call(("admin", "teacher"))
def delete_assignment(current_user, id):
    to_delete = _get_assignment(id, current_user)

    if current_user.account_type != "admin" and \
            to_delete.for_class not in current_user.classes:
        raise PermissionError(
            "You cannot delete an assignment for a class you're not teaching."
        )

    send_task(
        config["SISYPHUS_ADDRESS"],
        "delete_assignments",
        [str(to_delete.id)],
        ""
    )
    
    return (
        "%s has been queued for deletion. Please allow a few minutes for the "
        "task to complete." % _assignment_to_str(to_delete)
    )

@_api_call(("admin", "teacher"))
def get_archive(current_user, assignment, email = ""):
    # tar_tasks imports galahweb because it wants access to the logger, to
    # prevent a circular dependency we won't load the module until we need it.
    the_assignment = _get_assignment(assignment, current_user)

    if current_user.account_type != "admin" and \
            the_assignment.for_class not in current_user.classes:
        raise PermissionError(
            "You can only modify assignments for classes you teach."
        )

    # Create the task ID here rather than inside sisyphus so that we can tell
    # the user how to find the archive once its done.
    task_id = ObjectId()

    # We will not perform the work of archiving right now but will instead pass
    # if off to the heavy lifter to do it for us.
    send_task(
        config["SISYPHUS_ADDRESS"],
        "tar_bulk_submissions",
        str(task_id),
        current_user.email,
        str(the_assignment.id),
        email
    )
    
    return (
        "Your archive is being created.",
        {
            "X-Download": "archives/" + str(task_id),
            "X-Download-DefaultName": "submissions.tar.gz"
        }
    )

from types import FunctionType
api_calls = dict((k, v) for k, v in globals().items() if isinstance(v, APICall))

if __name__ == "__main__":
    import json
    print get_api_info("current_user")



# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

from galah.db.models import Assignment, Submission, Class, User
from bson import ObjectId
import os.path
import shutil

# Set up configuration and logging
from galah.base.config import load_config
config = load_config("sisyphus")

import logging
logger = logging.getLogger("galah.sisyphus.delete_assignments")

def _delete_assignments(ids, delete_class):
    if delete_class:
        delete_class = ObjectId(delete_class)

    # Convert all of the IDs we were given to ObjectIDs in one go
    ids = [ObjectId(i) for i in ids]

    logger.info("Deleting assignments %s.", str(ids))

    # Query the database for all the assignments we are supposed to delete
    assignments = list(Assignment.objects(id__in = ids))

    # Query the database for all the submissions we are supposed to delete, this
    # will potentially be an absolutely massive list, so we do not want to
    # place all the results into a list immediately like we did for the
    # assignments.
    submissions = Submission.objects(assignment__in = ids)

    # Delete assignments directory on the filesystem
    for i in assignments:
        try:
            shutil.rmtree(
                os.path.join(config["SUBMISSION_DIRECTORY"], str(i.id))
            )
        except:
            logger.error("Failed to delete assignment with ID %s", str(i.id))

            raise

    # Actually delete the submissions from the database
    Submission.objects(assignment__in = ids).delete()

    # Delete the assignments
    Assignment.objects(id__in = ids).delete()

    if delete_class:
        # Unenroll all students in the class
        User.objects(classes = delete_class).update(
            pull__classes = delete_class
        )

        # Delete the class
        Class.objects(id = delete_class).delete()


# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains a number of useful decorators and functions that are so
handy they're almost magical.

"""

import collections
import functools

# http://wiki.python.org/moin/PythonDecoratorLibrary#Memoize
class memoize(object):
   """
   Decorator. Caches a function's return value each time it is called.
   If called later with the same arguments, the cached value is returned
   (not reevaluated).

   """

   def __init__(self, func):
      self.func = func
      self.cache = {}

   def __call__(self, *args):
      if not isinstance(args, collections.Hashable):
         # uncacheable. a list, for instance.
         # better to not cache than blow up.
         return self.func(*args)

      if args in self.cache:
         return self.cache[args]
      else:
         value = self.func(*args)
         self.cache[args] = value
         return value

   def __repr__(self):
      "Return the function's docstring."

      return self.func.__doc__

   def __get__(self, obj, objtype):
      "Support instance methods."

      return functools.partial(self.__call__, obj)


from setuptools import setup, find_packages

setup(
    name = "galah",
    version = "0.1.3",
    packages = find_packages( 
        exclude = [
            "galah.sheep*", 
            "galah.shepherd*"
        ]
    ),

    include_package_data = True,
    zip_safe = False,
    
    entry_points = {
        "console_scripts": [
            "create_admin.py = utils.create_admin:main",
            "api_client.py = utils.api_client:main"
        ]
    },

    install_requires = [
        "Flask>=0.8", 
        "Flask-Login>=0.1.3", 
        "Flask-WTF>=0.6",
        "WTForms>=1.0.1",
        "mongoengine>=0.6.9",
        "decorator>=3.4.0",
        "requests>=1.0.3",
        "google-api-python-client>=1.0",
        "simple-pbkdf2>=1.0",
        "pyzmq>=2.2.0"
    ],

    package_data = {
        # If any package contains *.txt or *.rst files, include them:
        "web": ["*.html", "*.css"],
    },

    # metadata for upload to PyPI
    author = "John Sullivan",
    author_email = "jsull003@ucr.edu",
    description = "Galah is a utility to allow better professer student communication",
    license = "AGPLv3",
)



# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

from _login import login
from _browse_assignments import browse_assignments
from _view_assignment import view_assignment
from _upload_submission import upload_submission
from _logout import logout
from _error import error
from _home import home
from _api import api_login, api_call
from _get_archive import get_archive
from _download_submission import download_submission


# Copyright 2012-2013 John Sullivan
# Copyright 2012-2013 Other contributers as noted in the CONTRIBUTERS file
#
# This file is part of Galah.
#
# Galah is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Galah is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Galah.  If not, see <http://www.gnu.org/licenses/>.

## Create the form that takes simple archives ##
from flask.ext.wtf import Form, FieldList, FileField, validators, BooleanField

class SimpleArchiveForm(Form):
    archive = FieldList(FileField("Archive"), min_entries = 3)
    
## The Actual View ##
from galah.web import app
from flask.ext.login import current_user
from galah.web.auth import account_type_required
from bson.objectid import ObjectId
from bson.errors import InvalidId
from flask import abort, render_template, get_flashed_messages
from galah.db.models import Assignment, Submission
from galah.base.pretty import pretty_time
from galah.web.util import create_time_element, GalahWebAdapter
import datetime
import logging

logger = GalahWebAdapter(logging.getLogger("galah.web.views.view_assignment"))

@app.route("/assignments/<assignment_id>/")
@account_type_required(("student", "teacher"))
def view_assignment(assignment_id):
    simple_archive_form = SimpleArchiveForm()
    
    # Convert the assignment in the URL into an ObjectId
    try:
        id = ObjectId(assignment_id)
    except InvalidId:
        logger.info("Invalid Assignment ID requested.")
        
        abort(404)
    
    # Retrieve the assignment
    try:
        assignment = Assignment.objects.get(id = id)
    except Assignment.DoesNotExist:
        logger.info("Non-extant ID requested.")
        
        abort(404)

    # Get all of the submissions for this assignmnet
    submissions = list(
        Submission.objects(
            user = current_user.id,
            assignment = id
        ).order_by(
            "-most_recent",
            "-timestamp"
        )
    )
    
    # Add the pretty version of each submissions timestamp
    for i in submissions:
        i.timestamp_pretty = pretty_time(i.timestamp)
    
    return render_template(
        "assignment.html",
        now = datetime.datetime.today(),
        create_time_element = create_time_element,
        assignment = assignment, 
        submissions = submissions,
        simple_archive_form = simple_archive_form,
        new_submissions = [v for k, v in get_flashed_messages(with_categories = True) if k == "new_submission"]
    )
